from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.ext.associationproxy import association_proxy, _AssociationList
from sqlalchemy import (String, Unicode, Integer, Boolean, Index,
                        ForeignKey, ForeignKeyConstraint, UniqueConstraint)
from sqlalchemy.orm import exc, relationship
import os
from collections import OrderedDict
from .jmdict import (Base, CA,
                     StrColumnMixin, IdxStrColumnMixin,
                     KanjiCharacterAssociation)
from .kanjidic_txt import dicref_keys
from jikan_sqlalchemy_utils import repr_helper, index, OColumn

_dr_str = set(('moro', 'oneill_names', 'busy_people_chapter'))
dicref_types = OrderedDict((k, dict(
    coltype=(Integer if k not in _dr_str else Unicode(8)),
    func=(int if k not in _dr_str else str)))
                           for k in dicref_keys)

DicRefMixin = type('DicRefMixin', (), {
    'dr_'+k: OColumn(v['coltype'], _order=i+50)
    for i,(k,v) in enumerate(dicref_types.items())})

cp_keys = ('ucs', 'jis208', 'jis212', 'jis213')

class Character(CA, DicRefMixin, Base):
    __tablename__  = 'k_c'
    str = OColumn(Unicode(4), primary_key=True, _order=0)

    cp_ucs    = OColumn(Integer, _order=10)
    cp_jis208 = OColumn(Integer, _order=11)
    cp_jis212 = OColumn(Integer, _order=12)
    cp_jis213 = OColumn(Integer, _order=13)

    radicals    = relationship("Radical",   back_populates="character")
    variants    = relationship("Variant",   back_populates="character")
    querycodes  = relationship("QueryCode", back_populates="character")
    rms         = relationship("RMGroup",   back_populates="character")
    nanori_     = relationship("Nanori",    back_populates="character")
    nanori      = association_proxy('nanori_', 'str')
    wrong_stroke_counts_ = relationship(
        "WrongStrokeCount", back_populates="character")
    wrong_stroke_counts  = association_proxy(
        "wrong_stroke_counts_", "stroke_count")
    radical_names_ = relationship(
        "RadicalName", back_populates="character")
    radical_names  = association_proxy(
        "radical_names_", "str")


    grade        = OColumn(Integer, _order=20)
    stroke_count = OColumn(Integer, _order=21)
    freq         = OColumn(Integer, _order=22)
    rad_name     = OColumn(Unicode, _order=23)
    jlpt         = OColumn(Integer, _order=24)

    def __str__(self):
        return self.str

    def __repr__(self):
        return repr_helper('character', (
            (None, str(self)),
            ('cp_ucs', self.cp_ucs),
            ('cp_jis208', self.cp_jis208),
            ('cp_jis212', self.cp_jis212),
            ('cp_jis213', self.cp_jis213),
            ('grade', self.grade),
            ('stroke_count', self.stroke_count),
            ('freq', self.freq),
            ('rad_name', self.rad_name),
            ('jlpt', self.jlpt),
            ('radicals', self.radicals),
            ('variants', self.variants),
            ('querycodes', self.querycodes),
            ('rms', self.rms),
            ('nanori', self.nanori)))

    @declared_attr
    def kanji_entries(cls):
        return relationship(
            "KanjiEntry", secondary=KanjiCharacterAssociation.__table__,
            viewonly=True)

class CharacterChildMixin:
    @declared_attr
    def character_id(cls):
        return OColumn(Unicode(4), ForeignKey(Character.__tablename__+'.str'),
                       nullable=False, primary_key=True, _order=0)

    @declared_attr
    def i(cls):
        return OColumn(Integer,
                       nullable=False, primary_key=True, _order=1)

    @declared_attr
    def character(cls):
        return relationship("Character", back_populates=cls._back_populates)

class Radical(CA, CharacterChildMixin, Base):
    __tablename__ = 'k_radical'
    _back_populates = 'radicals'
    value = OColumn(Integer,     _order=10)
    type  = OColumn(Unicode(16), _order=11)

    def __repr__(self):
        return repr_helper('rad', (
            (None, str(self.value)),
            ('type', self.type)))

class Variant(CA, CharacterChildMixin, IdxStrColumnMixin, Base):
    __tablename__ = 'k_variant'
    _back_populates = 'variants'
    str  = OColumn(Unicode(16), _order=10)
    type = OColumn(Unicode(16), _order=11)

    def __repr__(self):
        return repr_helper('var', (
            (None, str(self)),
            ('type', self.type)))

class WrongStrokeCount(CA, CharacterChildMixin, Base):
    __tablename__ = 'k_wrong_stroke_count'
    _back_populates = 'wrong_stroke_counts_'
    stroke_count = OColumn(Integer, _order=10)

    def __repr__(self):
        return repr_helper('wsc', ((None, str(self.stroke_count)),))

class RadicalName(CA, CharacterChildMixin, IdxStrColumnMixin, Base):
    __tablename__ = 'k_rad_name'
    _back_populates = 'radical_names_'
    str = OColumn(Unicode(16), _order=10)

    def __repr__(self):
        return repr_helper('rad', ((None, str(self)),))

class QueryCode(CA, CharacterChildMixin, IdxStrColumnMixin, Base):
    __tablename__ = 'k_querycode'
    _back_populates = 'querycodes'
    str  = OColumn(Unicode(16), _order=10)
    type = OColumn(Unicode(16), _order=11)
    skip_misclass = OColumn(Unicode(16), _order=12)

    def __repr__(self):
        return repr_helper('qc', (
            (None, str(self)),
            ('type', self.type),
            ('skip_misclass', self.skip_misclass or False)))

class Nanori(CA, CharacterChildMixin, IdxStrColumnMixin, Base):
    __tablename__ = 'k_nanori'
    _back_populates = 'nanori_'
    str = OColumn(Unicode(16), _order=10)

    def __repr__(self):
        return repr_helper('nanori', (
            (None, str(self)),))

class RMGroup(CA, CharacterChildMixin, Base):
    __tablename__ = 'k_rm'
    _back_populates = 'rms'
    readings = relationship("Reading", back_populates="rm")
    meanings = relationship("Meaning", back_populates="rm")

    def __repr__(self):
        return repr_helper('rmgroup', (
            ('readings', self.readings),
            ('meanings', self.meanings)))

class RMGroupChildMixin:
    @declared_attr
    def __partial_table_args__(cls):
        t = RMGroup.__tablename__
        return (
            ForeignKeyConstraint(
                [   'character_id', 'rm_i'],
                [t+'.character_id', t+'.i']),)

    @declared_attr
    def character_id(cls):
        return OColumn(Unicode(4), nullable=False, primary_key=True, _order=0)

    @declared_attr
    def rm_i(cls):
        return OColumn(Integer,    nullable=False, primary_key=True, _order=1)

    @declared_attr
    def i(cls):
        return OColumn(Integer,    nullable=False, primary_key=True, _order=2)

    @declared_attr
    def rm(cls):
        return relationship("RMGroup", back_populates=cls._back_populates)

class Reading(CA, RMGroupChildMixin, IdxStrColumnMixin, Base):
    __tablename__ = 'k_r'
    _back_populates = 'readings'
    str  = OColumn(Unicode(64), _order=10)
    type = OColumn(Unicode(8),  _order=11)
    on_type = OColumn(Unicode(8), _order=14)
    jouyou  = OColumn(Boolean,    _order=15)

    def __repr__(self):
        return repr_helper('kr', (
            (None, str(self)),
            ('type', self.type),
            ('on_type', self.on_type or False),
            ('jouyou', self.jouyou or False)))

class Meaning(CA, RMGroupChildMixin, StrColumnMixin, Base):
    __tablename__ = 'k_m'
    _back_populates = 'meanings'
    str  = OColumn(Unicode(128), _order=10)
    lang = OColumn(Unicode(5),  _order=20)

    def __repr__(self):
        return repr_helper('km', (
            (None, str(self)),
            ('lang', self.lang)))




from ..main import kanjidic_parse, jmdict_parse, cross_reference
from ..processor.jmdict   import JMDictProcessor
from ..processor.kanjidic import KanjiDicProcessor
from jikan_sqlalchemy_utils.session import SessionMaker

from ..db.jmdict import DictionaryEntry, Base
from ..db.kanjidic import Character

import os
import unittest
import tempfile
from pkg_resources import resource_stream

class IncompleteTest(unittest.TestCase):
    def setUp(self):
        _, fn = tempfile.mkstemp(suffix='.jmdict_tmp')
        self.db_filename = fn
        self.sm = sm = SessionMaker.from_sqlite_filename(
            fn, isolation_level='SERIALIZABLE')

        sm.populate(Base)

        with resource_stream(__name__, 'kanjidic_sample.xml') as h:
            kanjidic_parse(sm(), h)

        with resource_stream(__name__, 'jmdict_sample.xml') as h:
            jmdict_parse(sm(), h)

        cross_reference(sm)

    def tearDown(self):
        os.remove(self.db_filename)

    def test_dict(self):
        q = self.sm().query
        tatoe = q(DictionaryEntry).get(1597120)

        k0 = tatoe.kanjis[0]
        r0 = tatoe.readings[0]

        self.assertEqual(k0.str, '例え')
        self.assertEqual(k0.pri_nf, 28)

        self.assertEqual(r0.str, 'たとえ')

        s0 = tatoe.senses[0]
        s0g0 = tatoe.senses[0].glosses[0]

        self.assertEqual(s0g0.str, 'example')
        self.assertEqual(s0g0.lang, 'eng')

        self.assertTrue('esp. 例え' in s0.infos)

    def test_kanji(self):
        q = self.sm().query
        rei = q(Character).get('例')
        rm0 = rei.rms[0]

        self.assertTrue(
            any(str(r)=='レイ' and r.type=='ja_on'
                for r in rm0.readings))
        self.assertTrue(
            any(str(m)=='example' and m.lang=='en'
                for m in rm0.meanings))

        self.assertTrue(set(str(ke) for ke in rei.kanji_entries)
                        .issuperset({'例え', '例外'}))





import lxml.etree as etree
from gzip import GzipFile
from .processor.jmdict   import JMDictProcessor
from .processor.kanjidic import KanjiDicProcessor
from jikan_sqlalchemy_utils.session import SessionMaker
from sys import stderr
from .db.jmdict import Base

def iterparse_clearing(filename_or_file):
    """Yield elements from *filename_or_file* xml incrementally, \
while clearing reference from the root node to its children to allow them \
to be garbage collected. This assumes that the xml file contains a \
single root node with many (direct) children."""
    # parser = etree.XMLParser(resolve_entities=False)
    context = iter(etree.iterparse(
        filename_or_file, events=('start', 'end'),
        resolve_entities=False))
    _, root = next(context) # get root element
    for event, elem in context:
        if event == 'end':
            yield elem
            if len(root):
                root.clear() # free memory

def yield_per(query, n=100):
    '''
like the `yield_per` method, but doesn't keep the cursor busy.
the SQL query must have a well-defined order'''
    i = 0
    while True:
        xs = tuple(query.slice(i, i+n).all())
        yield from xs
        if len(xs) < n:
            break
        i += n

def cross_reference(sm):
    from .db.kanjidic import Character
    from .db.jmdict import KanjiEntry, KanjiCharacterAssociation

    S = sm()
    q = S.query(KanjiCharacterAssociation)
    q.delete()
    S.commit()

    S = sm()
    to_be_saved = []
    def flush():
        S.bulk_save_objects(to_be_saved)
        S.commit()
        to_be_saved.clear()

    q = S.query(Character)

    qke = S.query(KanjiEntry).order_by(KanjiEntry.entry_id, KanjiEntry.i)
    for idx, ke in enumerate(yield_per(qke, n=10)):
        entry_id = ke.entry_id
        ke_i     = ke.i
        for c in frozenset(str(ke)):
            char = q.get(c)
            if char is not None:
                to_be_saved.append(KanjiCharacterAssociation(
                    entry_id=entry_id, ke_i=ke_i,
                    str=char.str))
        if len(to_be_saved) >= 1000:
            print("{:>7d}".format(idx), file=stderr)
            flush()
    flush()

def jmdict_parse(session, stream):
    p = JMDictProcessor(session=session)
    n = 0
    for e in iterparse_clearing(stream):
        if e.tag == 'entry':
            p.process_entry(e)
            n += 1
            if n % 500 == 0:
                print('{:>7d}'.format(n), file=stderr)
    p.flush(force=True)

def kanjidic_parse(session, stream):
    p = KanjiDicProcessor(session=session)
    n = 0
    for e in iterparse_clearing(stream):
        if e.tag == 'character':
            p.process_character(e)
            n += 1
            if n % 500 == 0:
                print('{:>7d}'.format(n), file=stderr)
    p.flush(force=True)

def main():
    import argparse
    parser = argparse.ArgumentParser(
        description=("Generate EDICT sqlite db."))
    parser.add_argument('--drop-all', action='store_true',
                        help='drop all pyjmdict tables')
    parser.add_argument('--jmdict',
                        help='path to EDICT "JMdict.xml.gz"')
    parser.add_argument('--kanjidic',
                        help='path to EDICT "kanjidic2.xml.gz"')
    parser.add_argument('--crossref', action='store_true',
                        help='cross-reference jmdict and kanjidic')
    parser.add_argument('-u', '--pyjmdict-db', required=True,
                        help='file from which to read database URI')
    args = parser.parse_args()

    sm = SessionMaker.from_uri_filename(
        args.pyjmdict_db, isolation_level='SERIALIZABLE')

    # drop all tables
    if args.drop_all:
        Base.metadata.drop_all(sm().bind)

    sm.populate(Base)

    if args.kanjidic:
        with open(args.kanjidic, 'rb') as compressed_h, \
             GzipFile(fileobj=compressed_h) as h:
            kanjidic_parse(sm(), h)

    if args.jmdict:
        with open(args.jmdict, 'rb') as compressed_h, \
             GzipFile(fileobj=compressed_h) as h:
            jmdict_parse(sm(), h)

    if args.crossref:
        cross_reference(sm.sessionmaker)

if __name__=='__main__':
    main()



from sqlalchemy import (String, Unicode, Integer, Boolean, Index,
                        ForeignKey, ForeignKeyConstraint, UniqueConstraint)

dicref_text = """
nelson_c - "Modern Reader's Japanese-English Character Dictionary",  
	edited by Andrew Nelson (now published as the "Classic" 
	Nelson).
nelson_n - "The New Nelson Japanese-English Character Dictionary", 
	edited by John Haig.
halpern_njecd - "New Japanese-English Character Dictionary", 
	edited by Jack Halpern.
halpern_kkd - "Kodansha Kanji Dictionary", (2nd Ed. of the NJECD)
	edited by Jack Halpern.
halpern_kkld - "Kanji Learners Dictionary" (Kodansha) edited by 
	Jack Halpern.
halpern_kkld_2ed - "Kanji Learners Dictionary" (Kodansha), 2nd edition
  (2013) edited by Jack Halpern.
heisig - "Remembering The  Kanji"  by  James Heisig.
heisig6 - "Remembering The  Kanji, Sixth Ed."  by  James Heisig.
gakken - "A  New Dictionary of Kanji Usage" (Gakken)
oneill_names - "Japanese Names", by P.G. O'Neill. 
oneill_kk - "Essential Kanji" by P.G. O'Neill.
moro - "Daikanwajiten" compiled by Morohashi. For some kanji two
	additional attributes are used: m_vol:  the volume of the
	dictionary in which the kanji is found, and m_page: the page
	number in the volume.
moro_vol - Volume of the dictionary in which the kanji is found.
moro_page - Page number in the volume.
henshall - "A Guide To Remembering Japanese Characters" by
	Kenneth G.  Henshall.
sh_kk - "Kanji and Kana" by Spahn and Hadamitzky.
sh_kk2 - "Kanji and Kana" by Spahn and Hadamitzky (2011 edition).
sakade - "A Guide To Reading and Writing Japanese" edited by
	Florence Sakade.
jf_cards - Japanese Kanji Flashcards, by Max Hodges and
      Tomoko Okazaki. (Series 1)
henshall3 - "A Guide To Reading and Writing Japanese" 3rd
      edition, edited by Henshall, Seeley and De Groot.
tutt_cards - Tuttle Kanji Cards, compiled by Alexander Kask.
crowley - "The Kanji Way to Japanese Language Power" by
	Dale Crowley.
kanji_in_context - "Kanji in Context" by Nishiguchi and Kono.
busy_people_volume - "Japanese For Busy People" vols I-III, published
      by the AJLT. The codes are the volume.chapter.
busy_people_chapter - See busy_people_volume.
kodansha_compact - the "Kodansha Compact Kanji Guide".
maniette - codes from Yves Maniette's "Les Kanjis dans la tete" French adaptation of Heisig.
"""

def extract_keys(text):
    return [line.split(' - ')[0]
            for line in text.replace('\t',' '*4).split('\n')
            if line and not line.startswith(' ')]

dicref_keys = extract_keys(dicref_text)




from setuptools import setup, find_packages
pkg = "pyjmdict"
ver = '1.0.0'
setup(
    name             = pkg,
    version          = ver,
    description      = "JMDict sqlalchemy interface",
    long_description = "Interface to JMDict and KanjiDic based on sqlalchemy",
    author           = "jikan@cock.li",
    author_email     = "jikan@cock.li",
    license          = "LGPLv3",
    url              = "https://hydra.ecd.space/jikan/pyjmdict/",
    zip_safe         = True,
    packages         = find_packages(),
    install_requires = ['sqlalchemy>=1', 'cachetools>=1', 'lxml>=3',
                        'jikan_sqlalchemy_utils>=1.0.0'],
    classifiers      = ["Programming Language :: Python :: 3 :: Only"])



from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.ext.associationproxy import association_proxy, _AssociationList
from sqlalchemy import (String, Unicode, Integer, Boolean, Index,
                        ForeignKey, ForeignKeyConstraint, UniqueConstraint)
from sqlalchemy.orm import exc, relationship
import os
from jikan_sqlalchemy_utils import (
    make_merge_args_class, OColumn, repr_helper, index)

Base = declarative_base()
CA = make_merge_args_class(Base)

# note that columns named "i" refer to the index of the child in its parent

class StrColumnMixin:
    _str_len = 8

    @declared_attr
    def str(cls):
        return OColumn(Unicode(cls._str_len), _order=10)

    def __str__(self):
        return self.str

class IdxStrColumnMixin(StrColumnMixin):
    @declared_attr
    def __partial_table_args__(cls):
        return (index(cls, 'str'),)

class DictionaryEntry(Base):
    __tablename__ = 'd_entry'

    id = OColumn(Integer, primary_key=True, _order=0) # ent_seq
    kanjis   = relationship("KanjiEntry"  , back_populates="parent_entry")
    readings = relationship("ReadingEntry", back_populates="parent_entry")
    senses   = relationship("Sense"       , back_populates="parent_entry")

    def __repr__(self):
        return repr_helper('E', (
            ('ent_seq', self.id),
            ('kanjis', self.kanjis),
            ('readings', self.readings),
            ('senses', self.senses)))

pri_keys = ['news', 'ichi', 'spec', 'gai', 'nf']

class PriorityMixin:
    pri_news = OColumn(Integer, nullable=True, _order=50)
    pri_ichi = OColumn(Integer, nullable=True, _order=51)
    pri_spec = OColumn(Integer, nullable=True, _order=52)
    pri_gai  = OColumn(Integer, nullable=True, _order=53)
    pri_nf   = OColumn(Integer, nullable=True, _order=54)

    @declared_attr
    def __partial_table_args__(cls):
        return (index(cls, 'pri_news'),
                index(cls, 'pri_ichi'),
                index(cls, 'pri_spec'),
                index(cls, 'pri_gai'),
                index(cls, 'pri_nf'))

    def _pri_kw(self):
        return tuple((k, getattr(self, 'pri_'+k))
                     for k in pri_keys)

class DictionaryEntryChildMixin:
    @declared_attr
    def entry_id(cls):
        return OColumn(Integer, ForeignKey("d_entry.id"),
                       nullable=False, primary_key=True, _order=0)
    @declared_attr
    def i(cls):
        return OColumn(Integer,
                       nullable=False, primary_key=True, _order=1)

class KanjiReadingAssociation(CA, Base): # re_restr
    __tablename__  = 'd_kr'

    @declared_attr
    def __partial_table_args__(cls):
        return (
            ForeignKeyConstraint(
                [     'entry_id',   'ke_i'],
                ['d_ke.entry_id', 'd_ke.i']),
            ForeignKeyConstraint(
                [     'entry_id',   're_i'],
                ['d_re.entry_id', 'd_re.i']),
            index(cls, 'er', 'entry_id', 're_i'),
            index(cls, 'ek', 'entry_id', 'ke_i'))

    entry_id = OColumn(Integer, primary_key=True, _order=0)
    ke_i     = OColumn(Integer, primary_key=True, _order=1)
    re_i     = OColumn(Integer, primary_key=True, _order=2)

    @declared_attr
    def ke(cls):
        return relationship("KanjiEntry", viewonly=True)
    @declared_attr
    def re(cls):
        return relationship("ReadingEntry", viewonly=True)

    def __repr__(self):
        return repr_helper('KR', (
            ('ent_seq', self.entry_id),
            ('ke_i', self.ke_i),
            ('re_i', self.re_i)))

class KanjiCharacterAssociation(CA, IdxStrColumnMixin, Base): # re_restr
    __tablename__  = 'dk_kc'

    @declared_attr
    def __partial_table_args__(cls):
        return (
            ForeignKeyConstraint(
                [     'entry_id',   'ke_i'],
                ['d_ke.entry_id', 'd_ke.i']),
            ForeignKeyConstraint(
                ['str'],
                ['k_c.str']),
            index(cls, 'ek', 'entry_id', 'ke_i'))

    entry_id = OColumn(Integer,    primary_key=True, _order=0)
    ke_i     = OColumn(Integer,    primary_key=True, _order=1)
    str      = OColumn(Unicode(4), primary_key=True, _order=2)

    @declared_attr
    def ke(cls):
        return relationship("KanjiEntry", viewonly=True)

    @declared_attr
    def character(cls):
        return relationship("Character", viewonly=True)

    def __repr__(self):
        return repr_helper('KC', (
            ('ent_seq', self.entry_id),
            ('ke_i', self.ke_i),
            ('c', self.str)))

class SenseKanji(CA, Base):
    __tablename__  = 'd_sk'

    @declared_attr
    def __partial_table_args__(cls):
        return (
            ForeignKeyConstraint(
                [     'entry_id',   'ke_i'],
                ['d_ke.entry_id', 'd_ke.i']),
            ForeignKeyConstraint(
                [     'entry_id',   'se_i'],
                ['d_se.entry_id', 'd_se.i']),
            index(cls, 'entry_id'))

    entry_id = OColumn(Integer, primary_key=True, _order=0)
    se_i     = OColumn(Integer, primary_key=True, _order=1)
    ke_i     = OColumn(Integer, primary_key=True, _order=2)

    @declared_attr
    def se(cls):
        return relationship("Sense", viewonly=True)
    @declared_attr
    def ke(cls):
        return relationship("KanjiEntry", viewonly=True)

class SenseReading(CA, Base):
    __tablename__  = 'd_sr'

    @declared_attr
    def __partial_table_args__(cls):
        return (
            ForeignKeyConstraint(
                [     'entry_id',   're_i'],
                ['d_re.entry_id', 'd_re.i']),
            ForeignKeyConstraint(
                [     'entry_id',   'se_i'],
                ['d_se.entry_id', 'd_se.i']),
            index(cls, 'entry_id'))

    entry_id = OColumn(Integer, primary_key=True, _order=0)
    se_i     = OColumn(Integer, primary_key=True, _order=1)
    re_i     = OColumn(Integer, primary_key=True, _order=2)

    @declared_attr
    def se(cls):
        return relationship("Sense", viewonly=True)
    @declared_attr
    def re(cls):
        return relationship("ReadingEntry", viewonly=True)

class KanjiEntry(CA, PriorityMixin, IdxStrColumnMixin,
                 DictionaryEntryChildMixin, Base):
    __tablename__  = 'd_ke'

    @declared_attr
    def parent_entry(cls):
        return relationship("DictionaryEntry", back_populates="kanjis")

    @declared_attr
    def readings(cls):
        return relationship(
            "ReadingEntry", secondary=KanjiReadingAssociation.__table__,
            viewonly=True)

    @declared_attr
    def senses(cls):
        return relationship(
            "Sense", secondary=SenseKanji.__table__,
            viewonly=True)

    @declared_attr
    def characters(cls):
        return relationship(
            "Character", secondary=KanjiCharacterAssociation.__table__,
            viewonly=True)

    str = OColumn(Unicode(32), _order=10)

    def __repr__(self):
        return repr_helper('K', ((None, self.str),) + tuple(self._pri_kw()))

class ReadingEntry(CA, PriorityMixin, IdxStrColumnMixin,
                   DictionaryEntryChildMixin, Base):
    __tablename__  = 'd_re'

    @declared_attr
    def parent_entry(cls):
        return relationship("DictionaryEntry", back_populates="readings")

    @declared_attr
    def kanjis(cls):
        return relationship(
            "KanjiEntry", secondary=KanjiReadingAssociation.__table__,
            viewonly=True)

    @declared_attr
    def senses(cls):
        return relationship(
            "Sense", secondary=SenseReading.__table__,
            viewonly=True)

    str          = OColumn(Unicode(64), _order=10)
    true_reading = OColumn(Boolean,     _order=11)

    def __repr__(self):
        return repr_helper('R', ((None, self.str),) + tuple(self._pri_kw()) + (
            ('not_true_reading', not self.true_reading),))

def _assoc_proxy_str(attr):
    return association_proxy(attr+'_', 'str')

def _sense_rel_helper(name):
    k = "Sense"+name
    return relationship(k, back_populates="sense", order_by=k+'.i')

class Sense(CA, DictionaryEntryChildMixin, Base):
    __tablename__ = 'd_se'

    parent_entry = relationship("DictionaryEntry", back_populates="senses")

    xrefs_           = _sense_rel_helper("XRef")
    antonyms_        = _sense_rel_helper("Antonym")
    parts_of_speech_ = _sense_rel_helper("POS")
    fields_          = _sense_rel_helper("Field")
    miscs_           = _sense_rel_helper("Misc")
    lsources         = _sense_rel_helper("LSource")
    dialects_        = _sense_rel_helper("Dialect")
    glosses          = _sense_rel_helper("Gloss")
    infos_           = _sense_rel_helper("Info")

    xrefs           = _assoc_proxy_str('xrefs')
    antonyms        = _assoc_proxy_str('antonyms')
    infos           = _assoc_proxy_str('infos')
    parts_of_speech = _assoc_proxy_str('parts_of_speech')
    fields          = _assoc_proxy_str('fields')
    miscs           = _assoc_proxy_str('miscs')
    dialects        = _assoc_proxy_str('dialects')

    @declared_attr
    def kanjis(cls):
        return relationship(
            "KanjiEntry", secondary=SenseKanji.__table__,
            viewonly=True)

    @declared_attr
    def readings(cls):
        return relationship(
            "ReadingEntry", secondary=SenseReading.__table__,
            viewonly=True)

    def __repr__(self):
        kw = (('pos', 'parts_of_speech'),
              ('fields', None),
              ('miscs', None),
              ('dial', 'dialects'),
              ('infos', None),
              ('glosses', None),
              ('lsrc', 'lsources'),
              ('xrefs', None),
              ('ant', 'antonyms'))
        return repr_helper(
            'sense', ((k, getattr(self, v if v else k)) for k,v in kw),
            ignore_empty=True)

class SenseChildMixin:
    @declared_attr
    def __partial_table_args__(cls):
        return (
            ForeignKeyConstraint(
                [     'entry_id',   'se_i'],
                ['d_se.entry_id', 'd_se.i']),)

    entry_id = OColumn(Integer, primary_key=True, _order=0)
    se_i     = OColumn(Integer, primary_key=True, _order=1)
    i        = OColumn(Integer, primary_key=True, _order=2)

    def __str__(self):
        return self.str

    def __repr__(self):
        return "<{} {!r}>".format(
            self.__class__.__name__, str(self))

class SenseXRef(CA, IdxStrColumnMixin, SenseChildMixin, Base):
    __tablename__ = 'd_se_xref'
    _tagname = 'xref'
    _str_len = 32*4

    @declared_attr
    def sense(cls):
        return relationship("Sense", back_populates="xrefs_")

class SenseAntonym(CA, IdxStrColumnMixin, SenseChildMixin, Base):
    __tablename__ = 'd_se_antonym'
    _tagname   = 'ant'
    _str_len = 16

    @declared_attr
    def sense(cls):
        return relationship("Sense", back_populates="antonyms_")

class SensePOS(CA, IdxStrColumnMixin, SenseChildMixin, Base):
    '''part of speech'''
    __tablename__ = 'd_se_pos'
    _tagname   = 'pos'
    _tagentity = True
    _str_len = 16

    @declared_attr
    def sense(cls):
        return relationship("Sense", back_populates="parts_of_speech_")

class SenseField(CA, IdxStrColumnMixin, SenseChildMixin, Base):
    '''field of application of sense'''
    __tablename__ = 'd_se_field'
    _tagname = 'field'
    _tagentity = True
    _str_len = 12

    @declared_attr
    def sense(cls):
        return relationship("Sense", back_populates="fields_")

class SenseMisc(CA, IdxStrColumnMixin, SenseChildMixin, Base):
    '''misc information for sense'''
    __tablename__ = 'd_se_misc'
    _tagname = 'misc'
    _tagentity = True
    _str_len = 8

    @declared_attr
    def sense(cls):
        return relationship("Sense", back_populates="miscs_")

class SenseLSource(CA, StrColumnMixin, SenseChildMixin, Base):
    '''loanword source'''
    __tablename__ = 'd_se_lsource'

    str     = OColumn(Unicode(256), _order=10)
    lang    = OColumn(Unicode(5),   _order=20)
    partial = OColumn(Boolean,      _order=21) # ls_type
    wasei   = OColumn(Boolean,      _order=22) # ls_wasei

    @declared_attr
    def sense(cls):
        return relationship("Sense", back_populates="lsources")

    def __repr__(self):
        return repr_helper('lsource', (
            (None, self.str)
            ('lang', self.lang or False),
            ('partial', self.partial),
            ('wasei', self.wasei)))

class SenseDialect(CA, IdxStrColumnMixin, SenseChildMixin, Base):
    '''dialect information for sense'''
    __tablename__ = 'd_se_dial'
    _tagname = 'dial'
    _tagentity = True
    _str_len = 8

    @declared_attr
    def sense(cls):
        return relationship("Sense", back_populates="dialects_")

class SenseGloss(CA, StrColumnMixin, SenseChildMixin, Base):
    __tablename__ = 'd_se_gloss'

    @declared_attr
    def __partial_table_args__(cls):
        return (index(cls, 'lang'),
                index(cls, 'gend'))

    str  = OColumn(Unicode(1024), _order=10)
    lang = OColumn(Unicode(   5), _order=20)
    gend = OColumn(Unicode(   5), _order=21, nullable=True)

    @declared_attr
    def sense(cls):
        return relationship("Sense", back_populates="glosses")

    def __repr__(self):
        return repr_helper('gloss', (
            (None, self.str),
            ('gend', self.gend or False),
            ('lang', self.lang or False)))

class SenseInfo(CA, IdxStrColumnMixin, SenseChildMixin, Base):
    __tablename__ = 'd_se_info'
    _tagname    = 's_inf'
    _str_len = 512

    @declared_attr
    def sense(cls):
        return relationship("Sense", back_populates="infos_")

class KanjiInf(CA, IdxStrColumnMixin, Base):
    __tablename__ = 'd_ke_inf'
    _str_len = 8

    @declared_attr
    def __partial_table_args__(cls):
        return (
            ForeignKeyConstraint(
                [     'entry_id',   'ke_i'],
                ['d_ke.entry_id', 'd_ke.i']))

    entry_id = OColumn(Integer, primary_key=True, _order=0)
    ke_i     = OColumn(Integer, primary_key=True, _order=1)
    i        = OColumn(Integer, primary_key=True, _order=2)

    def __repr__(self):
        return repr_helper('ke_inf', (
            (None, str(self)),
            ('ent_seq', self.entry_id),
            ('ke_i', self.ke_i),
            ('i', self.i)))

class ReadingInf(CA, IdxStrColumnMixin, Base):
    __tablename__ = 'd_re_inf'
    _str_len = 8

    @declared_attr
    def __partial_table_args__(cls):
        return (
            ForeignKeyConstraint(
                [     'entry_id',   're_i'],
                ['d_re.entry_id', 'd_re.i']),)

    entry_id = OColumn(Integer, primary_key=True, _order=0)
    re_i     = OColumn(Integer, primary_key=True, _order=1)
    i        = OColumn(Integer, primary_key=True, _order=2)

    def __repr__(self):
        return repr_helper('re_inf', (
            (None, str(self)),
            ('ent_seq', self.entry_id),
            ('re_i', self.ke_i),
            ('i', self.i)))
